

import numpy as np
from .spencoding import normalize


def aggregate_neighbors(data, graph, mode='add', weighted=False, norm='l1', order=1):
    if mode not in ['append', 'add']:
        raise Exception('Only `append` or `add` methods are accepted')

    ndata = np.zeros_like(data)
    neighbors = np.zeros(data.shape[0])

    for node in graph:
        nh = graph.neighbors(node)
        neighbors[node] = len(nh)
        if not weighted:
            ndata[node] = data[nh].sum(axis=0)
        else:
            weights = np.array([graph[node][nb]['boundary'] for nb in nh], float)
            ndata[node] = (data[nh] * weights[:, None]).sum(axis=0)

        if order == 2:
            currents = nh + [node]
            for nhnode in nh:
                nhnh = filter(lambda x: None if x in currents else x,
                              graph.neighbors(nhnode))
                if len(nhnh) == 0:
                    continue
                currents += nhnh
                neighbors[node] += len(nhnh)
                if not weighted:
                    ndata[node] += data[nhnh].sum(axis=0)
                else:
                    weights = np.array([graph[nhnode][nb]['boundary'] for nb in nhnh], float)
                    ndata[node] += (data[nhnh] * weights[:, None]).sum(axis=0)

    if mode == 'add':
        ndata += data
        neighbors += 1

    if norm == 'mean':
        ndata /= neighbors[:, None]
    else:
        ndata = normalize(ndata, norm=norm)

    if mode == 'append':
        ndata = np.c_[data, ndata]

    return ndata





from ..qt_compat import QtGui, QtCore, QtWidgets

from .base import HWidgets, HeaderLabel, FileWidget, CheckableCombo, \
                  ComboDialog

from .mpl_widgets import MplCanvas

from ..core import DataModel, Launcher

from ..actions import volread

import numpy as np
import os

class LoadWidget(QtWidgets.QWidget):

    data_loaded = QtCore.pyqtSignal(str)

    def __init__(self, parent=None):
        super(LoadWidget, self).__init__(parent)

        self.DM = DataModel.instance()
        self.launcher = Launcher.instance()
        self.data = None

        main_layout = QtWidgets.QVBoxLayout()
        main_layout.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter)
        self.setLayout(main_layout)

        container = QtWidgets.QWidget(self)
        hbox = QtWidgets.QHBoxLayout(container)
        container.setMaximumWidth(950)
        container.setMaximumHeight(530)
        container.setLayout(hbox)
        container.setObjectName("loaderContainer")
        container.setStyleSheet('QWidget#loaderContainer {'
                                '  background-color: #fefefe; '
                                '  border-radius: 10px;'
                                '}')
        lvbox = QtWidgets.QVBoxLayout()
        rvbox = QtWidgets.QVBoxLayout()
        lvbox.setAlignment(QtCore.Qt.AlignTop)
        rvbox.setAlignment(QtCore.Qt.AlignTop)
        hbox.addLayout(lvbox, 1)
        hbox.addLayout(rvbox, 1)

        main_layout.addWidget(container)

        ################################################
        # LEFT
        ################################################
        lvbox.addWidget(HeaderLabel('Preview Dataset'))

        self.wperspective = QtWidgets.QComboBox()
        self.wperspective.addItem('Axial')
        self.wperspective.addItem('Coronal')
        self.wperspective.addItem('Sagittal')

        self.winvert = CheckableCombo()
        self.winvert.addItem('Invert Z')
        self.winvert.addItem('Invert Y')
        self.winvert.addItem('Invert X')
        self.winvert.addItem('Transpose X <-> Y')

        self.slider = QtWidgets.QSlider(1)
        lvbox.addWidget(self.slider)

        lvbox.addWidget(HWidgets('3D Perspective:', self.wperspective,
                                 'Axis Order:', self.winvert,
                                 stretch=[0,1,0,1]))

        self.canvas = MplCanvas()
        lvbox.addWidget(self.canvas)

        ################################################
        # RIGHT
        ################################################

        # INPUT
        rvbox.addWidget(HeaderLabel('Input Dataset'))

        self.winput = FileWidget(extensions='*.mrc *.rec *.npy *.h5 *.hdf5 *.tif *.tiff', save=False)
        rvbox.addWidget(HWidgets('Select File:', self.winput, stretch=[0,1]))

        # OUTPUT
        rvbox.addWidget(HeaderLabel('Workspace Folder'))
        self.woutput = FileWidget(folder=True)

        rvbox.addWidget(HWidgets('Select Folder:', self.woutput, stretch=[0,1]))
        rvbox.addWidget(QtWidgets.QWidget(), 1)

        # Save | Cancel
        self.cancel = QtWidgets.QPushButton('Cancel')
        self.cancel.setMinimumWidth(100)
        self.cancel.setMinimumHeight(70)
        self.load = QtWidgets.QPushButton('Load')
        self.load.setMinimumWidth(100)
        self.load.setMinimumHeight(70)
        rvbox.addWidget(HWidgets(None, self.cancel, self.load, stretch=[1,0,0]))

        ###########################

        self.winput.path_updated.connect(self.load_data)
        self.wperspective.currentIndexChanged.connect(self.on_perspective)
        self.winvert.selectionChanged.connect(self.on_axis)
        self.slider.valueChanged.connect(self.update_image)
        self.load.clicked.connect(self.on_load_data)

    def load_data(self, path):
        if path is not None and len(path) > 0:
            dataset = None
            if path.endswith('.h5') or path.endswith('.hdf5'):
                available_hdf5 = self.DM.available_hdf5_datasets(path)
                selected, accepted = ComboDialog.getOption(available_hdf5, parent=self)
                if accepted == QtWidgets.QDialog.Rejected:
                    return
                dataset = selected

            self.launcher.run(volread, path=path, dataset=dataset,
                              caption='Loading data...',
                              cb=self.on_data_loaded)

    def on_data_loaded(self, result):
        data, vmin, vmax, evmin, evmax = result
        if data is None or data.ndim != 3:
            raise Exception('SuRVoS can only process 3 dimensional data: {}'.format(data.shape))

        self.DM.tmp_data = data
        self.DM.tmp_stats = (vmin, vmax, evmin, evmax)
        self.update_image(idx=None)

    def transform_data(self):
        tidx = self.wperspective.currentIndex()
        trans = [0,1,2]
        tchanged = False
        if tidx != 0:
            trans = [1,0,2] if tidx == 1 else [2,0,1]
            tchanged = True

        inv_indexes = self.winvert.getSelectedIndexes()
        if 3 in inv_indexes:
            trans[1], trans[2] = trans[2], trans[1]
            tchanged = True

        if tchanged:
            data = np.transpose(self.DM.tmp_data, trans)
        else:
            data = self.DM.tmp_data

        if 0 in inv_indexes:
            data = data[::-1]
        if 1 in inv_indexes:
            data = data[:, ::-1]
        if 2 in inv_indexes:
            data = data[:, :, ::-1]

        return data

    def on_axis(self, *args):
        self.update_image()

    def on_perspective(self, *args):
        self.update_image()

    def update_image(self, idx=None):
        if idx is None:
            self.data = self.transform_data()
            idx = self.data.shape[0]//2
            self.slider.blockSignals(True)
            self.slider.setMinimum(0)
            self.slider.setMaximum(self.data.shape[0] - 1)
            self.slider.setValue(idx)
            self.slider.blockSignals(False)
            self.canvas.ax.set_ylim([self.data.shape[1] + 1, -1])
            self.canvas.ax.set_xlim([-1, self.data.shape[2] + 1])

        evmin, evmax = self.DM.tmp_stats[-2:]

        img = self.data[idx]
        self.canvas.ax.imshow(img, 'gray', vmin=evmin, vmax=evmax)
        self.canvas.ax.grid(False)
        self.canvas.redraw()

    def on_load_data(self):
        wspath = self.woutput.value()
        if self.data is not None and wspath is not None:
            if not os.path.isdir(wspath):
                errmsg = 'Workspace Path not valid: {}'.format(wspath)
                QtWidgets.QMessageBox.critical(self, 'Error', errmsg)
                return

            do_continue = True
            if len(os.listdir(wspath)) > 0:
                errmsg = 'Workspace folder is not empty, do you still want to continue? It might overwrite and remove previous files.'
                ans = QtWidgets.QMessageBox.question(self, "Workspace Folder not empty",
                                                 errmsg,
                                                 QtWidgets.QMessageBox.Yes,
                                                 QtWidgets.QMessageBox.No)
                if ans == QtWidgets.QMessageBox.No:
                    do_continue = False

            if do_continue:
                self.DM.tmp_data = np.ascontiguousarray(self.data)
                del self.data
                self.data_loaded.emit(wspath)
        else:
            if self.data is None:
                errmsg = 'Data has to be loaded first'
                QtWidgets.QMessageBox.critical(self, 'Error', errmsg)
                return
            if wspath is None:
                errmsg = 'Workspace path has to be set first'
                QtWidgets.QMessageBox.critical(self, 'Error', errmsg)
                return




from ..qt_compat import QtGui, QtCore

import numpy as np
import logging as log

import os
import glob
import h5py as h5

from scipy.stats import gaussian_kde

from .singleton import Singleton


@Singleton
class DataModel(QtCore.QObject):

    roi_changed = QtCore.pyqtSignal()
    select_channel = QtCore.pyqtSignal(str)
    update_channel = QtCore.pyqtSignal(str)
    cropped = QtCore.pyqtSignal()
    vmin_changed = QtCore.pyqtSignal(float)
    vmax_changed = QtCore.pyqtSignal(float)
    evmin_changed = QtCore.pyqtSignal(float)
    evmax_changed = QtCore.pyqtSignal(float)
    channel_computed = QtCore.pyqtSignal(str, dict)
    channel_removed = QtCore.pyqtSignal(str)
    feature_updated = QtCore.pyqtSignal(str)
    voxel_descriptor_computed = QtCore.pyqtSignal()
    voxel_descriptor_removed = QtCore.pyqtSignal()
    supervoxel_descriptor_computed = QtCore.pyqtSignal()
    supervoxel_descriptor_removed = QtCore.pyqtSignal()
    supervoxel_descriptor_cleared = QtCore.pyqtSignal()
    level_predicted = QtCore.pyqtSignal(int)

    Axial = 0
    Sagittal = 1
    Coronal = 2

    def __init__(self):
        QtCore.QObject.__init__(self)

        self.main_window = None

        # tmp_data
        self.tmp_data = None
        self.tmp_stats = None

        # Storage
        self.wspath = None
        self.current_idx = 0
        self.selected_gpu = -1

        # DataFiles
        self.data = None
        self.data_shape = None
        self.active_roi = slice(None), slice(None), slice(None)

        # Superpixels
        self.svlabels = None
        self.svindex = None
        self.svtable = None
        self.svtotal = None

        # Megavoxels
        self.mvlabels = None
        self.mvindex = None
        self.mvtable = None
        self.mvtotal = None

        # GT Labels
        self.gtlevel = 0
        self.gtradius = 1
        self.gtinterpolation = 'linear'
        self.gtselected = None
        self.last_changes = None
        self.growing_bbox = ['data', 1., [10, 50, 50], 0]

        # KDE plots
        self.computed_histograms = dict()
        self.channel_computed.connect(self.disable_histogram)

    def disable_histogram(self, name):
        if name in self.computed_histograms:
            del self.computed_kde[name]

    def get_histogram(self, dataset):
        if dataset in self.computed_histograms:
            return self.computed_histograms[dataset]

        log.info('+ Loading channel into memory')
        data = self.load_ds(dataset).ravel()
        attr = self.attrs(dataset)
        amin = attr['vmin']
        amax = attr['vmax']

        log.info('+ Computing Histogram')
        y, x = np.histogram(data[~np.isnan(data)], 10000)
        x = (x[:-1] + x[1:]) / 2.

        self.computed_histograms[dataset] = (x, y)
        return (x, y)

    ##########################################################################
    # TOI Transformation
    ##########################################################################

    def region_shape(self):
        return tuple(map(lambda x: x.stop - x.start, self.active_roi))

    def transform_slices(self, slice_z=None, slice_y=None, slice_x=None):
        slices = []

        for t, s in zip(self.active_roi, (slice_z, slice_y, slice_x)):
            start = t.start
            stop = t.stop

            try:
                s = int(s)
                start += s
                stop = start + 1
            except:
                if s is None:
                    s = slice(None)
                if s.stop is not None:
                    stop = t.start + s.stop
                if s.start is not None:
                    start += s.start

            slices.append(slice(start, stop))

        return slices

    ##########################################################################
    # DATASETS IN / OUT
    ##########################################################################

    def create_empty_dataset(self, dataset, shape, dtype, check=True,
                             params=None, fillvalue=-1):
        ds_file = self.ds_path(dataset)
        ds_folder = os.path.dirname(ds_file)
        if not os.path.exists(ds_folder):
            os.mkdir(ds_folder)
        if check and os.path.isfile(ds_file):
            return False
        with h5.File(ds_file, 'w') as f:
            ds = f.create_dataset('data', shape=shape, fillvalue=fillvalue,
                                  dtype=dtype)
            if params is not None:
                for k, v in params.items():
                    ds.attrs[k] = v
        return True

    def write_dataset(self, dataset, data, params=None):
        ds_file = self.ds_path(dataset)
        ds_folder = os.path.dirname(ds_file)
        if not os.path.exists(ds_folder):
            os.mkdir(ds_folder)
        with h5.File(ds_file, 'a') as f:
            if 'data' in f:
                f['data'].write_direct(data)
            else:
                f.create_dataset('data', data=data)
            if params is not None:
                for k, v in params.items():
                    f['data'].attrs[k] = v

    def load_ds(self, dataset):
        ds_file = self.ds_path(dataset)
        with h5.File(ds_file, 'r') as f:
            data = np.zeros_like(f['data'])
            f['data'].read_direct(data)
        return data

    def load_slices(self, dataset, slice_z=None, slice_y=None, slice_x=None,
                    apply_roi=True):
        ds_file = self.ds_path(dataset)
        if apply_roi:
            sz, sy, sx = self.transform_slices(slice_z, slice_y, slice_x)
        else:
            sz, sy, sx = slice_z, slice_y, slice_x
        with h5.File(ds_file, 'r') as f:
            data = f['data'][sz, sy, sx]
        return data

    def write_slices(self, dataset, data, slice_z=None, slice_y=None,
                     slice_x=None, params=None, apply_roi=True):
        ds_file = self.ds_path(dataset)
        if apply_roi:
            sz, sy, sx = self.transform_slices(slice_z, slice_y, slice_x)
        else:
            sz, sy, sx = slice_z, slice_y, slice_x
        with h5.File(ds_file, 'a') as f:
            f['data'][sz, sy, sx] = data
            if params is not None:
                for k, v in params.items():
                    f['data'].attrs[k] = v
        return data

    def remove_dataset(self, dataset):
        ds_file = self.ds_path(dataset)
        if os.path.exists(ds_file):
            os.remove(ds_file)

    ##########################################################################
    # ATTRIBUTES AND MEMBERSHIP
    ##########################################################################

    def set_attrs(self, dataset, attrs):
        ds_file = self.ds_path(dataset)
        with h5.File(ds_file, 'a') as f:
            for k, v in attrs.items():
                if isinstance(v, list): #this helps with python 3 compatibility
                    if len(v)!=0:
                        if isinstance(v[0], str) or isinstance(v[0], bytes):
                            v = [np.string_(x) for x in v]
                f['data'].attrs[k] = v

    def attrs(self, dataset):
        ds_file = self.ds_path(dataset)
        with h5.File(ds_file, 'r') as f:
            attr = dict(f['data'].attrs)
        return attr

    def attr(self, dataset, attr, value=None):
        ds_file = self.ds_path(dataset)
        with h5.File(ds_file, 'r') as f:
            if attr in f['data'].attrs:
                if value is None:
                    val = f['data'].attrs[attr]
                else:
                    val = (f['data'].attrs[attr] == value)
            else:
                val = None
        return val

    def ds_shape(self, dataset):
        ds_file = self.ds_path(dataset)
        with h5.File(ds_file, 'r') as f:
            shape = f['data'].shape
        return shape

    def has_ds(self, dataset):
        active = False
        ds_file = self.ds_path(dataset)
        if os.path.isfile(ds_file):
            with h5.File(ds_file, 'r') as f:
                if 'data' in f and 'active' in f['data'].attrs:
                    active = f['data'].attrs['active']
        return active

    def has_grp(self, group):
        ds_folder = self.grp_path(group)
        return os.path.exists(ds_folder)

    def ds_path(self, dataset):
        if dataset.startswith('/'):
            dataset = dataset[1:]
        return os.path.join(self.wspath, '{}.h5'.format(dataset))

    def grp_path(self, group):
        return os.path.join(self.wspath, group)

    ##########################################################################
    # SCAN UTILITIES
    ##########################################################################

    def scan_datasets_group(self, group, shape=None, dtype=None, path=''):
        datasets = []
        for name, ds in group.items():
            curr_path = '{}/{}'.format(path, name)
            if hasattr(ds, 'shape'):
                if len(ds.shape) == 3 \
                        and (shape is None or ds.shape == shape) \
                        and (dtype is None or ds.dtype == dtype):
                    datasets.append(curr_path)
            else:
                extra = self.scan_datasets_group(ds, shape=shape, path=curr_path)
                if len(extra) > 0:
                    datasets += extra
        return datasets


    def available_hdf5_datasets(self, path, shape=None, dtype=None):
        datasets = []
        with h5.File(path, 'r') as f:
            datasets = self.scan_datasets_group(f, shape=shape, dtype=dtype)
        return datasets

    ##########################################################################
    # FEATURES AND DESCRIPTORS
    ##########################################################################

    def find_dataset(self, group, filter_active=True):
        grp_dir = self.grp_path(group)
        ds = glob.glob('{}/*.h5'.format(grp_dir))
        ds = ['{}/{}'.format(group, os.path.basename(d[:-3])) for d in ds]
        if filter_active:
            ds = [d for d in ds if self.attr(d, 'active') == True]
        return ds

    def available_channels(self, return_names=False, filter_active=True):
        if self.has_grp('channels'):
            possible = self.find_dataset('channels', filter_active=filter_active)
            af = []
            for f in possible:
                attrs = self.attrs(f)
                if 'active' in attrs and 'feature_idx' in attrs:
                    fidx = attrs['feature_idx']
                    if return_names:
                        af.append((fidx, f, attrs))
                    else:
                        af.append((fidx, f))
        else:
            af = []
        return af

    def available_annotations(self, return_names=False):
        if self.has_grp('annotations'):
            possible = self.find_dataset('annotations')
            af = []
            for f in possible:
                attrs = self.attrs(f)
                if 'active' in attrs and 'levelid' in attrs:
                    af.append((attrs['levelid'], f, attrs['active']))
        else:
            af = []
        return af





from ..qt_compat import QtGui, QtCore, QtWidgets

from .base import HWidgets, HeaderLabel

class RefContainer(QtWidgets.QWidget):

    def __init__(self, title, link_url, description, parent=None):
        super(RefContainer, self).__init__(parent)
        self.setFixedWidth(250)
        self.setFixedHeight(150)
        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)
        title = QtWidgets.QLabel('<a href="{}">{}</a>'.format(link_url, title))
        title.setAlignment(QtCore.Qt.AlignTop | QtCore.Qt.AlignHCenter)
        title.setStyleSheet('font-size: 18pt; color: #6194BC;')
        title.setOpenExternalLinks(True)
        vbox.addWidget(title)
        desc = QtWidgets.QLabel(description)
        desc.setWordWrap(True)
        desc.setAlignment(QtCore.Qt.AlignTop | QtCore.Qt.AlignHCenter)
        desc.setStyleSheet('font-size: 14pt;')
        vbox.addWidget(desc, 1)

class PreButton(QtWidgets.QPushButton):

    def __init__(self, title, description, parent=None):
        super(PreButton, self).__init__(parent)
        self.setStyleSheet('QPushButton {'
                           '  background-color: #6194BC; color: #fefefe;'
                           '}'
                           'QPushButton:hover {'
                           '  background-color: #009999;'
                           '}')
        self.setFixedWidth(250)
        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)
        title = QtWidgets.QLabel('{}'.format(title))
        title.setAlignment(QtCore.Qt.AlignTop | QtCore.Qt.AlignHCenter)
        title.setStyleSheet('font-size: 18pt; border-radius: 2px;')
        vbox.addWidget(title)

        desc = QtWidgets.QLabel(description)
        desc.setWordWrap(True)
        desc.setAlignment(QtCore.Qt.AlignTop | QtCore.Qt.AlignHCenter)
        desc.setStyleSheet('background-color: transparent; font-size: 14pt; color: #fefefe;')
        vbox.addWidget(desc, 1)
        title.mousePressEvent = self.mousePressEvent
        desc.mousePressEvent = self.mousePressEvent

class PreWidget(QtWidgets.QWidget):

    def __init__(self, parent=None):
        super(PreWidget, self).__init__(parent)
        main_layout = QtWidgets.QVBoxLayout()
        main_layout.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter)
        self.setLayout(main_layout)

        container = QtWidgets.QWidget(self)
        vbox = QtWidgets.QVBoxLayout(container)
        container.setMaximumWidth(950)
        container.setMaximumHeight(530)
        container.setLayout(vbox)
        container.setStyleSheet('QWidget {'
                                '  background-color: #fefefe; '
                                '  border-radius: 10px;'
                                '}')
        main_layout.addWidget(container)

        title = HeaderLabel('<b>SuRVoS</b>: <u>Su</u>per-<u>R</u>egion '
                            '<u>Vo</u>lume <u>S</u>egmentation workbench',
                            fontsize=28, padding=30, margin=5)
        vbox.addWidget(title)

        ref1 = RefContainer('Source Repository',
                            'https://github.com/DiamondLightSource/SuRVoS',
                            'Find the latest version of the software, contribute or suggest improvements.')
        ref2 = RefContainer('Documentation',
                            'https://github.com/DiamondLightSource/SuRVoS/wiki',
                            'Discover how does <b>SuRVoS</b> work and how to get the'
                            ' most out of it.')
        ref3 = RefContainer('Issues and Help',
                            'https://github.com/DiamondLightSource/SuRVoS/issues',
                            'Did you have any trouble or did you find any bug? '
                            'We will try to help.')
        vbox.addWidget(HWidgets(None, ref1, None, ref2, None, ref3, None,
                                stretch=[1,0,1,0,1,0,1]))

        self.open = PreButton('Open Dataset',
                              'Load an existing dataset of supported file formats:'
                              '\n\nMRC (.mrc, .rec), HDF5 (.h5, .hdf5)'
                              ', Tiff Stacks (.tif, .tiff)', parent=self)
        self.open.setFixedWidth(300)
        self.open.setFixedHeight(250)
        self.load = PreButton('Load workspace',
                              'Load a workspace previously created with SuRVoS.'
                              '\n\nAll the feature channels, super-regions and annotations'
                              ' will be recovered.', parent=self)
        self.load.setFixedWidth(300)
        self.load.setFixedHeight(250)

        actions_container = QtWidgets.QWidget()
        hbox = QtWidgets.QHBoxLayout()
        actions_container.setLayout(hbox)

        hbox.addWidget(QtWidgets.QWidget(), 1)
        hbox.addWidget(self.open)
        hbox.addWidget(QtWidgets.QWidget(), 1)
        hbox.addWidget(self.load)
        hbox.addWidget(QtWidgets.QWidget(), 1)

        vbox.addWidget(actions_container)
        vbox.addWidget(QtWidgets.QWidget(), 1)




import os
import h5py as h5
import numpy as np
import copy
import tifffile

from ..lib.io import MRC

from ..core import DataModel

import logging as log

def volread(path=None, dataset='data', stats=True, normalize=True):
    file_name, file_extension = os.path.splitext(path)
    data = None

    log.info('+ Loading data to memory')

    if file_extension in ['.rec', '.mrc']:
        data = np.copy(MRC(path).data)
    elif file_extension == '.npy':
        data = np.load(path)
    elif file_extension in ['.hdf5', '.h5']:
        with h5.File(path, 'r') as f:
            data = f[dataset][:]
    elif file_extension in ['.tif', '.tiff']:
        data = tifffile.imread(path, multifile=True)
    else:
        raise Exception('File format not supported')

    if normalize:
        log.info('+ Normalizing data')
        data = data.astype(np.float32)
        mean = data.mean()
        std = data.std()
        data -= mean
        data /= std

    if stats:
        log.info('+ Calculating data statistics')
        vmin, vmax = data.min(), data.max()
        evmin, evmax = np.percentile(data, (1., 99.))

        return data, vmin, vmax, evmin, evmax
    else:
        return data


def load_data(data=None, stats=None):
    vmin, vmax, evmin, evmax = stats
    log.info('+ Writing data to disk')
    DM = DataModel.instance()
    params = dict(vmin=vmin, vmax=vmax, evmin=evmin, evmax=evmax,
                  default_evmin=evmin, default_evmax=evmax,
                  cmap='gray', active=True)
    DM.write_dataset('/data', data, params=params)





from __future__ import division

import numpy as np
import logging as log

from sklearn.metrics import cohen_kappa_score, jaccard_similarity_score

from ..core import DataModel

DM = DataModel.instance()


def compare_segmentations(levelA=None, levelB=None, labelsA=None, labelsB=None,
                          dice=False, jacc=False, cohen=False):
    log.info('+ Loading level A into memory')
    levelA = DM.load_slices(levelA)

    indexes = set()
    result = dict()

    log.info('+ Mapping labels from A')
    mappingA = np.full(levelA.shape, -1, np.int8)
    for s, t in labelsA.items():
        mappingA[levelA[...] == s] = t
        indexes |= {t}

    log.info('+ Loading level B into memory')
    levelB = DM.load_slices(levelB)

    log.info('+ Mapping labels from B')
    mappingB = np.full(levelB.shape, -1, np.int8)
    for s, t in labelsB.items():
        mappingB[levelB[...] == s] = t
        indexes |= {t}

    mappingA += 1
    mappingB += 1

    log.info('+ Computing intersection and overlapping')
    overlap = ((mappingA == mappingB) & (mappingA > 0))
    union = ((mappingA > 0) | (mappingB > 0))

    n = max(indexes) + 2
    m = float(mappingA.size)

    log.info('+ Couting target labels')
    inter = (mappingA * overlap).ravel()
    result['countsOverlap'] = np.bincount(inter, minlength=n) / m
    result['countsA'] = np.bincount(mappingA.ravel(), minlength=n) / m
    result['countsB'] = np.bincount(mappingB.ravel(), minlength=n) / m

    log.info('+ Computing scores')
    ninter = overlap.sum()
    nnzA = (mappingA > 0).sum()
    nnzB = (mappingB > 0).sum()

    dA = mappingA[union]
    dB = mappingB[union]

    if dice:
        result['dice'] = (2 * ninter) / float(nnzA + nnzB)
    if jacc:
        result['jacc'] = jaccard_similarity_score(dB.ravel(), dA.ravel())
    if cohen:
        result['cohen'] = cohen_kappa_score(mappingB.ravel(), mappingA.ravel(),
                                            labels=[i + 1 for i in indexes])

    result['indexes'] = indexes | {-1}

    return result




import os
import numpy as np
from Cython.Distutils import build_ext
from sys import platform
import sys
import distutils
from distutils.spawn import spawn, find_executable
from distutils.debug import DEBUG
from distutils.errors import DistutilsPlatformError, DistutilsExecError
PATH = os.environ.get('PATH')

#adapted fom http://code.activestate.com/recipes/52224-find-a-file-given-a-search-path/

def find_in_path(name, path):
    "Find a file in a search path"
    for cdir in path.split(os.pathsep):
        binpath = os.path.join(cdir, name)
        if os.path.exists(binpath):
            return os.path.abspath(binpath)

    return None

def locate_cuda():
    nvcc = None
    nvcc_exe = None
    nvcc_lib_path = None
    if platform == "win32":
        nvcc_exe = 'nvcc.exe'
        nvcc_lib_path = 'lib\\x64'
    else:
        nvcc_exe = 'nvcc'
        nvcc_lib_path = 'lib64'

    if 'CUDAHOME' in os.environ:
        home = os.environ['CUDAHOME']
        nvcc_path = os.path.join(home, 'bin', nvcc_exe)
        if os.path.isfile(nvcc_path):
            nvcc = nvcc_path

    if nvcc is None:
        nvcc_path = find_in_path(nvcc_exe, os.environ['PATH'])
        if nvcc_path is None or not os.path.isfile(nvcc_path):
            raise EnvironmentError('The nvcc binary could not be located in your'
                                   ' $PATH. Either add it to your path, or set'
                                   ' appropiate $CUDAHOME.')
        nvcc = nvcc_path
        home = os.path.dirname(os.path.dirname(nvcc_path))

    cudaconfig = {
        'home': home,
        'nvcc': nvcc,
        'include': os.path.join(home, 'include'),
        'lib64': os.path.join(home, nvcc_lib_path)
    }

    errmsg = 'The CUDA {} path could not be located in {}'

    for k, v in cudaconfig.items():
        if not os.path.exists(v):
            raise EnvironmentError(errmsg.format(k, v))

    return cudaconfig

CUDA = locate_cuda()

def customize_compiler_for_nvcc_unix(self):
    self.src_extensions.append('.cu')

    default_compiler_so = self.compiler_so
    super = self._compile

    def _compile(obj, src, ext, cc_args, extra_postargs, pp_opts):
        ext = os.path.splitext(src)[1]
        if ext == '.cu':
            self.set_executable('compiler_so', CUDA['nvcc'])
            postargs = extra_postargs['nvcc']
        elif type(extra_postargs) == dict:
            if ext == '.cpp':
                postargs = extra_postargs['g++']
            else:
                postargs = extra_postargs['gcc']
        else:
            postargs = extra_postargs

        super(obj, src, ext, cc_args, postargs, pp_opts)

        self.compiler_so = default_compiler_so

    self._compile = _compile

def customize_compiler_for_nvcc_win32(self):
    self.src_extensions.append('.cu')
    self.initialize()
    default_compiler_so = self.cc
    self.cc = CUDA['nvcc']
    self._cpp_extensions.append('.cu')

# Run the customize_compiler
class custom_build_ext(build_ext):

    def build_extensions(self):
        if platform == 'win32':
            customize_compiler_for_nvcc_win32(self.compiler)
            self.compiler.spawn = self.spawn
        else:
            customize_compiler_for_nvcc_unix(self.compiler)
        build_ext.build_extensions(self)


    def spawn(self, cmd, search_path=1, verbose=1, dry_run=0):
        """
        Perform any CUDA specific customizations before actually launching
        compile/link etc. commands.
        """
        print("Invoking custom spawn")
        if (sys.platform == 'darwin' and len(cmd) >= 2 and cmd[0] == 'nvcc' and
                cmd[1] == '--shared' and cmd.count('-arch') > 0):
            # Versions of distutils on OSX earlier than 2.7.9 inject
            # '-arch x86_64' which we need to strip while using nvcc for
            # linking
            while True:
                try:
                    index = cmd.index('-arch')
                    del cmd[index:index+2]
                except ValueError:
                    break
        elif self.compiler.compiler_type == 'msvc':
            # There are several things we need to do to change the commands
            # issued by MSVCCompiler into one that works with nvcc. In the end,
            # it might have been easier to write our own CCompiler class for
            # nvcc, as we're only interested in creating a shared library to
            # load with ctypes, not in creating an importable Python extension.
            # - First, we replace the cl.exe or link.exe call with an nvcc
            #   call. In case we're running Anaconda, we search cl.exe in the
            #   original search path we captured further above -- Anaconda
            #   inserts a MSVC version into PATH that is too old for nvcc.
            cmd[:1] = ['nvcc', '--compiler-bindir',
                       '"'+os.path.dirname(find_executable("cl.exe", PATH))+'"'
                       or cmd[0]]
            print(cmd)
            # - Secondly, we fix a bunch of command line arguments.
            for idx, c in enumerate(cmd):
                # create .dll instead of .pyd files
                #if '.pyd' in c: cmd[idx] = c = c.replace('.pyd', '.dll')
                # replace /c by -c
                if c == '/c': cmd[idx] = '-c'
                # replace /DLL by --shared
                elif c == '/DLL': cmd[idx] = '--shared'
                # remove --compiler-options=-fPIC
                elif '-fPIC' in c: del cmd[idx]
                # replace /Tc... by ...
                elif c.startswith('/Tc'): cmd[idx] = c[3:]
                elif c.startswith('/Tp'): cmd[idx] = c[3:]
                elif c.startswith('/MD'): cmd[idx] = '/MT'
                elif c.startswith('/MANIFEST:EMBED,ID=2'): cmd[idx] = ' '
                # replace /Fo... by -o ...
                elif c.startswith('/Fo'): cmd[idx:idx+1] = ['-o ', c[3:]]
                # replace /LIBPATH:... by -L...
                elif c.startswith('/LIBPATH:'): cmd[idx] = '-L' + c[9:]
                # replace /OUT:... by -o ...
                elif c.startswith('/OUT:'): cmd[idx:idx+1] = ['-o', c[5:]]
                # remove /EXPORT:initlibcudamat or /EXPORT:initlibcudalearn
                elif c.startswith('/EXPORT:'): del cmd[idx]
                # replace cublas.lib by -lcublas
                elif c == 'cublas.lib': cmd[idx] = '-lcublas'
            
            # - Finally, we pass on all arguments starting with a '/' to the
            #   compiler or linker, and have nvcc handle all other arguments
            if '--shared' in cmd:
                pass_on = '--cudart=shared --linker-options='
                # we only need MSVCRT for a .dll, remove CMT if it sneaks in:
                #cmd.append('/NODEFAULTLIB:libcmt.lib')
            else:
                pass_on = '--compiler-options='
            cmd = ([c for c in cmd if c[0] != '/'] +
                   [pass_on + ','.join(c for c in cmd if c[0] == '/')])
            # For the future: Apart from the wrongly set PATH by Anaconda, it
            # would suffice to run the following for compilation on Windows:
            # nvcc -c -O -o <file>.obj <file>.cu
            # And the following for linking:
            # nvcc --shared -o <file>.dll <file1>.obj <file2>.obj -lcublas
            # This could be done by a NVCCCompiler class for all platforms.
        self.spawn_nt(cmd, search_path, verbose, dry_run)
        
    
    def spawn_nt(self, cmd, search_path=1, verbose=0, dry_run=0):
        executable = cmd[0]
        if search_path:
            # either we find one or it stays the same
            executable = find_executable(executable) or executable
        print(' '.join([executable] + cmd[1:]))
        if not dry_run:
            # spawn for NT requires a full path to the .exe
            try:
                rc = os.spawnv(os.P_WAIT, executable, cmd)
            except OSError as exc:
                # this seems to happen when the command isn't found
                if not DEBUG:
                    cmd = executable
                raise DistutilsExecError(
                      "command %r failed: %s" % (cmd, exc.args[-1]))
            if rc != 0:
                # and this reflects the command running but failing
                if not DEBUG:
                    cmd = executable
                raise DistutilsExecError(
                      "command %r failed with exit status %d" % (cmd, rc))



import numpy as np
from ..qt_compat import QtGui, QtCore, QtWidgets

import logging as log
from collections import defaultdict

from matplotlib.colors import ListedColormap

from skimage.segmentation import find_boundaries

from ..core import Launcher
from ..widgets import LCheckBox, HWidgets, BLabel, PLineEdit, TComboBox, \
                      HeaderLabel, SubHeaderLabel, CSlider, ColorButton, \
                      HSize3D, MultiSourceCombo, ActionButton
from ..widgets.conficence_viewer import ConfidenceViewer
from .base import Plugin
from ..core import DataModel, LayerManager, LabelManager
from .. import actions as ac


class EnsembleWidget(QtWidgets.QWidget):

    predict = QtCore.pyqtSignal(dict)

    def __init__(self, parent=None):
        super(EnsembleWidget, self).__init__(parent=parent)

        vbox = QtWidgets.QVBoxLayout()
        vbox.setContentsMargins(0,0,0,0)
        self.setLayout(vbox)

        self.type_combo = TComboBox('Ensemble Type:', [
                                        'Random Forest',
                                        'ExtraRandom Forest',
                                        'AdaBoost',
                                        'GradientBoosting'
                                    ])
        self.type_combo.currentIndexChanged.connect(self.on_ensemble_changed)
        vbox.addWidget(self.type_combo)

        self.ntrees = PLineEdit(100, parse=int)
        self.depth = PLineEdit(None, parse=int)
        self.lrate = PLineEdit(1., parse=float)
        self.subsample = PLineEdit(1., parse=float)

        vbox.addWidget(HWidgets(QtWidgets.QLabel('# Trees:'),
                                self.ntrees,
                                QtWidgets.QLabel('Max Depth:'),
                                self.depth,
                                stretch=[0,1,0,1]))

        vbox.addWidget(HWidgets(QtWidgets.QLabel('Learn Rate:'),
                                self.lrate,
                                QtWidgets.QLabel('Subsample:'),
                                self.subsample,
                                stretch=[0,1,0,1]))

        self.btn_predict = ActionButton('Predict')
        self.btn_predict.clicked.connect(self.on_predict_clicked)
        self.n_jobs = PLineEdit(1, parse=int)
        vbox.addWidget(HWidgets('Num Jobs', self.n_jobs, None, self.btn_predict,
                                stretch=[0, 0,1,0]))

    def on_ensemble_changed(self, idx):
        if idx == 2:
            self.ntrees.setDefault(50)
        else:
            self.ntrees.setDefault(100)

        if idx == 3:
            self.lrate.setDefault(0.1)
            self.depth.setDefault(3)
        else:
            self.lrate.setDefault(1.)
            self.depth.setDefault(None)

    def on_predict_clicked(self):
        ttype = ['rf', 'erf', 'ada', 'gbf']
        params = {
            'clf'           : 'ensemble',
            'type'          : ttype[self.type_combo.currentIndex()],
            'n_estimators'  : self.ntrees.value(),
            'max_depth'     : self.depth.value(),
            'learning_rate' : self.lrate.value(),
            'subsample'     : self.subsample.value(),
            'n_jobs'        : self.n_jobs.value()
        }
        self.predict.emit(params)


class SVMWidget(QtWidgets.QWidget):

    predict = QtCore.pyqtSignal(dict)

    def __init__(self, parent=None):
        super(SVMWidget, self).__init__(parent=parent)

        vbox = QtWidgets.QVBoxLayout()
        vbox.setContentsMargins(0,0,0,0)
        self.setLayout(vbox)

        self.type_combo = TComboBox('Kernel Type:', [
                                        'linear',
                                        'poly',
                                        'rbf',
                                        'sigmoid'
                                    ])
        vbox.addWidget(self.type_combo)

        self.penaltyc = PLineEdit(1.0, parse=float)
        self.gamma = PLineEdit('auto', parse=float)

        vbox.addWidget(HWidgets(QtWidgets.QLabel('Penalty C:'),
                                self.penaltyc,
                                QtWidgets.QLabel('Gamma:'),
                                self.gamma,
                                stretch=[0,1,0,1]))

        self.btn_predict = ActionButton('Predict')
        self.btn_predict.clicked.connect(self.on_predict_clicked)
        vbox.addWidget(HWidgets(None, self.btn_predict, stretch=[1,0]))

    def on_predict_clicked(self):
        params = {
            'clf'       : 'svm',
            'kernel'    : self.type_combo.currentText(),
            'C'         : self.penaltyc.value(),
            'gamma'     : self.gamma.value()
        }

        self.predict.emit(params)


class OnlineWidget(QtWidgets.QWidget):

    predict = QtCore.pyqtSignal(dict)

    def __init__(self, parent=None):
        super(OnlineWidget, self).__init__(parent=parent)

        vbox = QtWidgets.QVBoxLayout()
        vbox.setContentsMargins(0,0,0,0)
        self.setLayout(vbox)

        self.type_combo = TComboBox('Loss function:', [
                                        #'hinge',
                                        'log',
                                        'modified_huber',
                                        #'squared_hinge',
                                        #'perceptron'
                                    ])
        vbox.addWidget(self.type_combo)

        self.penalty = TComboBox('Regularization:', [
                                        'none',
                                        'l1',
                                        'l2',
                                        'elasticnet',
                                    ])
        vbox.addWidget(self.penalty)

        self.alpha = PLineEdit(0.0001, parse=float)
        self.n_iter = PLineEdit(5, parse=int)

        vbox.addWidget(HWidgets(QtWidgets.QLabel('alpha:'),
                                self.alpha,
                                QtWidgets.QLabel('Num Iter:'),
                                self.n_iter,
                                stretch=[0,1,0,1]))

        self.chunks = PLineEdit(None, parse=int)

        self.btn_predict = ActionButton('Predict')
        self.btn_predict.clicked.connect(self.on_predict_clicked)
        vbox.addWidget(HWidgets(QtWidgets.QLabel('Chunks:'),
                                self.chunks,
                                None, self.btn_predict,
                                stretch=[0,0,1,0]))

    def on_predict_clicked(self):
        params = {
            'clf'       : 'sgd',
            'loss'      : self.type_combo.currentText(),
            'penalty'   : self.penalty.currentText(),
            'alpha'     : self.alpha.value(),
            'n_iter'    : self.n_iter.value(),
            'chunks'    : self.chunks.value()
        }

        self.predict.emit(params)


class TrainPredict(QtWidgets.QWidget):

    def __init__(self, parent=None):
        super(TrainPredict, self).__init__(parent=None)

        self.DM = DataModel.instance()
        self.LM = LayerManager.instance()
        self.LBLM = LabelManager.instance()
        self.launcher = Launcher.instance()

        self.selected_level = -1
        self.parent_level = None
        self.parent_label = -1

        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)

        vbox.addWidget(SubHeaderLabel('Descriptor'))

        # Region Type
        self.use_region = TComboBox('Region:', ['Voxels', 'SuperVoxels'], selected=1)
        self.use_region.currentIndexChanged.connect(self.on_region_changed)
        self.use_region.setMinimumWidth(200)

        # Descriptor type
        self.use_desc = MultiSourceCombo()

        self.desc_type = TComboBox('Type:',
                                    ['Mean', 'Quantized', 'Textons',
                                     'Covar', 'Sigma Set'])
        self.desc_bins = PLineEdit(10, parse=int)
        self.desc_bins.setMaximumWidth(60)

        self.nh_order = TComboBox('Order:', [0, 1, 2], selected=0)

        self.preprocess = TComboBox('Project:',
                                    ['None', 'standard', 'random_projection',
                                     'pca', 'random_pca'])

        vbox.addWidget(HWidgets(None, self.use_region, stretch=[1,0]))
        vbox.addWidget(HWidgets(None, 'Features:', self.use_desc, stretch=[1,0,0]))
        self.supervoxel_desc_params = HWidgets(self.desc_type, 'Bins:',
                                               self.desc_bins, self.nh_order,
                                               stretch=[1,0,0,1])
        vbox.addWidget(self.supervoxel_desc_params)
        vbox.addWidget(HWidgets(None, self.preprocess, stretch=[1,0,0]))

        vbox.addWidget(SubHeaderLabel('Classifier'))

        self.train_alg = TComboBox('Classifier type:', [
                                        'Ensemble',
                                        'SVM',
                                        'Online Linear'
                                   ])
        self.train_alg.currentIndexChanged.connect(self.on_classifier_changed)
        vbox.addWidget(self.train_alg)

        self.clf_container = QtWidgets.QWidget()
        vbox2 = QtWidgets.QVBoxLayout()
        vbox2.setContentsMargins(0,0,0,0)
        self.clf_container.setLayout(vbox2)

        self.ensembles = EnsembleWidget()
        self.ensembles.predict.connect(self.on_predict)
        self.svm = SVMWidget()
        self.svm.predict.connect(self.on_predict)
        self.online = OnlineWidget()
        self.online.predict.connect(self.on_predict)

        self.clf_container.layout().addWidget(self.ensembles)
        vbox.addWidget(self.clf_container)

        vbox.addWidget(SubHeaderLabel('Prediction Refinement'))
        self.refinement_combo = TComboBox('Refine:', ['None', 'Potts',
                                                      'Appearance', 'Edge'])
        self.refinement_lamda = PLineEdit(50., parse=float)
        self.refinement_lamda.setMaximumWidth(80)
        self.refinement_desc = MultiSourceCombo()
        self.ref_features = HWidgets(None, 'Features:', self.refinement_desc,
                                     stretch=[1,0,0])
        self.ref_features.setVisible(False)
        vbox.addWidget(HWidgets(self.refinement_combo, 'Lambda:', self.refinement_lamda))
        vbox.addWidget(self.ref_features)

        self.refinement_combo.currentIndexChanged.connect(self.on_ref_type_changed)
        self.use_region.currentIndexChanged.connect(self.on_region_changed)
        self.DM.voxel_descriptor_computed.connect(self.on_voxel_desc_changed)
        self.DM.voxel_descriptor_removed.connect(self.on_voxel_desc_changed)
        self.DM.supervoxel_descriptor_computed.connect(self.on_voxel_desc_changed)
        self.DM.supervoxel_descriptor_computed.connect(self.on_voxel_desc_changed)
        self.DM.supervoxel_descriptor_cleared.connect(self.on_voxel_desc_changed)

    def select_level(self, level, plevel, plabel):
        self.selected_level = level
        self.parent_level = plevel
        self.parent_label = plabel

    def on_voxel_desc_changed(self):
        self.on_region_changed(self.use_region.currentIndex())

    def on_ref_type_changed(self, idx):
        self.ref_features.setVisible(idx > 2)

    def on_classifier_changed(self, idx):
        if idx == 0:
            self.clf_container.layout().addWidget(self.ensembles)
            self.svm.setParent(None)
            self.online.setParent(None)
        elif idx == 1:
            self.clf_container.layout().addWidget(self.svm)
            self.ensembles.setParent(None)
            self.online.setParent(None)
        else:
            self.clf_container.layout().addWidget(self.online)
            self.svm.setParent(None)
            self.ensembles.setParent(None)

    def on_region_changed(self, idx):
        self.supervoxel_desc_params.setVisible(idx == 1)

    def on_predict(self, params):
        ### LEVEL AND LABEL
        level = self.selected_level

        if level < 0 or level is None:
            QtWidgets.QMessageBox.critical(self, "Error", "No level selected")
            return

        parent_level = self.parent_level
        parent_label = self.parent_label
        idxs = self.LBLM.idxs(level=level)

        if len(idxs) == 0:
            self.launcher.error.emit('Current level doesn\'t contain any label')
            return

        level_params = dict(level=level, plevel=parent_level, plabel=parent_label)

        ### DESCRIPTOR
        desc_params = dict()
        desc_params['supervoxels'] = 'supervoxels/supervoxels' if self.use_region.currentIndex() == 1 else None
        desc_params['features'] = self.use_desc.value()
        desc_params['projection'] = self.preprocess.value()
        if desc_params['supervoxels']:
            desc_params['desc_type'] = self.desc_type.value()
            desc_params['desc_bins'] = self.desc_bins.value()
            desc_params['nh_order'] = self.nh_order.value()
            desc_params['sp_edges'] = 'supervoxels/graph_edges'

        if len(desc_params['features']) == 0:
            self.launcher.error.emit('No descriptor feature has been selected.')
            return

        ### CLASSIFIER
        clf_params = params

        ### REFINEMENT
        ref_params = dict()
        ref_params['ref_type'] = self.refinement_combo.value()
        ref_params['lambda'] = self.refinement_lamda.value()
        ref_params['features'] = self.refinement_desc.value()
        ref_params['sp_edges'] = 'supervoxels/graph_edges'
        ref_params['sp_eweights'] = 'supervoxels/graph_edge_weights'

        if self.refinement_combo.currentIndex() > 2 and len(ref_params['features']) == 0:
            self.launcher.error.emit('No refinement feature has been selected.')
            return

        ### ANNOTATIONS
        y_data = self.LBLM.dataset(level)
        out_labels = 'predictions/predictions'
        out_confidence = 'predictions/confidence'

        if parent_level != None and parent_level >= 0 and parent_label >= 0:
            p_data = self.LBLM.dataset(parent_level)
        else:
            p_data = None

        self.DM.remove_dataset(out_labels)
        self.DM.remove_dataset(out_confidence)

        dataset = 'predictions/predictions'
        self.DM.create_empty_dataset(out_labels, self.DM.data_shape,
                                     np.int16, check=False)
        self.DM.create_empty_dataset(out_confidence, self.DM.data_shape,
                                     np.float32, check=False)

        log.info("Level {}".format(level_params))
        log.info("Desc {}".format(desc_params))
        log.info("Clf {}".format(clf_params))
        log.info("Ref {}".format(ref_params))
        log.info("{} {}".format(y_data, p_data))
        log.info("{} {}".format(out_labels, out_confidence))

        self.launcher.run(ac.predict_proba, y_data=y_data, p_data=p_data,
                          level_params=level_params, desc_params=desc_params,
                          clf_params=clf_params, ref_params=ref_params,
                          out_labels=out_labels, out_confidence=out_confidence,
                          cb=self.on_predicted,
                          caption='Predicting labels for Level {}'.format(level))

    def on_predicted(self, params):
        predictions, uncertainty, labels = params
        if not self.LM.isin('Predictions', 'Predictions'):
            self.LM.addLayer(predictions, 'Predictions', 'Predictions', alpha=0.5, visible=True)
        self.DM.level_predicted.emit(self.selected_level)


class UncertainLabelWidget(QtWidgets.QWidget):

    save = QtCore.pyqtSignal(int)

    def __init__(self, idx, name, color, parent=None):
        super(UncertainLabelWidget, self).__init__(parent=parent)

        self.idx = idx

        hbox = QtWidgets.QHBoxLayout()
        hbox.setContentsMargins(0, 0, 0, 0)
        self.setLayout(hbox)

        self.name = BLabel(name)
        self.name.setMinimumWidth(120)
        hbox.addWidget(self.name)

        self.color = ColorButton(color, clickable=False)
        hbox.addWidget(self.color, 1)

        self.apply = ActionButton('Save')
        hbox.addWidget(self.apply)

        self.apply.clicked.connect(self.on_save_label)

    def on_save_label(self):
        self.save.emit(self.idx)


class Uncertainty(QtWidgets.QWidget):

    def __init__(self, parent=None):
        super(Uncertainty, self).__init__(parent=parent)

        self.DM = DataModel.instance()
        self.LM = LayerManager.instance()
        self.LBLM = LabelManager.instance()

        self.selected_level = -1
        self.parent_level = None
        self.parent_label = -1

        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)

        self.level_header = SubHeaderLabel('No Level Predicted Yet')
        vbox.addWidget(self.level_header)

        self.combo_viz = TComboBox('Visualization:', ['Predictions', 'Confidence'])
        vbox.addWidget(self.combo_viz)

        self.labels = {}

        self.slider_thresh = CSlider('none', current=0)
        vbox.addWidget(HWidgets('Confidence:', self.slider_thresh, stretch=[0,1]))

        self.txt_from = HSize3D('From', default=(0, 0, 0), txtwidth=50)
        self.txt_to = HSize3D('To', default=self.DM.data_shape, txtwidth=50)
        vbox.addWidget(self.txt_from)
        vbox.addWidget(self.txt_to)

        self.combo_viz.currentIndexChanged.connect(self.on_viz_changed)
        self.slider_thresh.valueChanged.connect(self.on_thresh_changed)
        self.DM.level_predicted.connect(self.on_level_predicted)


    def on_level_predicted(self, idx):
        self.select_level(idx)

    def on_viz_changed(self, idx):
        if not self.LM.isin('Predictions', 'Predictions'):
            return

        labels_name = 'predictions/predictions'
        confidence_name = 'predictions/confidence'
        if idx == 0 and self.DM.attr(labels_name, 'active') == True:
            data = labels_name
            labels = self.DM.attr(labels_name, 'labels')
            colors = self.LBLM.colors(self.selected_level)
            idxs = self.LBLM.idxs(self.selected_level)
            maxid = 0 if len(idxs) == 0 else max(idxs)+1
            cmaplist = ['#000000'] * maxid
            for idx, color in zip(idxs, colors):
                cmaplist[idx] = color
            cmap = ListedColormap(cmaplist)
        else:
            data = confidence_name
            cmap = 'viridis'
            maxid = 1

        layer = self.LM.get('Predictions', 'Predictions')
        layer.data = data
        layer.cmap = cmap
        layer.vmin = 0
        layer.vmax = maxid

        if self.combo_viz.currentIndex() == 1:
            layer.threshold = None
        else:
            layer.threshold = (confidence_name, self.slider_thresh.value() / 100.)

        self.LM.setVisible('Predictions', 'Predictions', True)
        self.LM.update()

    def on_thresh_changed(self, thresh):
        self.on_viz_changed(self.combo_viz.currentIndex())


    def select_level(self, level):
        self.level_header.setText('Confidence Tool for Level {}'.format(level))
        self.selected_level = level

        for labelobj in self.labels.values():
            labelobj.setParent(None)

        self.labels = {}

        if level >= 0:
            for label in self.LBLM.labels(level):
                labelobj = UncertainLabelWidget(label.idx, label.name, label.color)
                labelobj.save.connect(self.on_save_label)
                self.layout().addWidget(labelobj)
                self.labels[label.idx] = labelobj

        self.on_viz_changed(self.combo_viz.currentIndex())

    def on_save_label(self, idx):
        if not self.LM.isin('Predictions', 'Predictions'):
            QtWidgets.QMessageBox.critical(self, 'Error', 'No label predicted')
            return

        f = self.txt_from.value()
        t = self.txt_to.value()
        s1 = slice(f[0], t[0])
        s2 = slice(f[1], t[1])
        s3 = slice(f[2], t[2])

        labels_name = 'predictions/predictions'
        uncertainty_name = 'predictions/confidence'
        threshold = self.slider_thresh.value() / 100.

        launcher = Launcher.instance()
        launcher.setup('Saving predictions for [{}]'.format(self.selected_level))
        log.info('+ Loading data into memory')
        pred = self.DM.load_slices(labels_name)
        conf = self.DM.load_slices(uncertainty_name)

        log.info('+ Masking annotations')
        mask = np.zeros(pred.shape, dtype=np.bool)
        np.logical_and((pred[s1, s2, s3] == idx), (conf[s1, s2, s3] >= threshold),
                       out=mask[s1, s2, s3])

        log.info('+ Saving results to disk')
        dataset = self.LBLM.dataset(self.selected_level)
        target = self.DM.load_slices(dataset)
        target[mask] = idx
        self.DM.write_slices(dataset, target)
        self.LM.update()

        launcher.cleanup()


class Training(Plugin):

    name = 'Model training'

    def __init__(self, parent=None):
        super(Training, self).__init__(ptype=Plugin.Plugin, parent=parent)

        self.DM = DataModel.instance()
        self.LM = LayerManager.instance()
        self.LBLM = LabelManager.instance()
        self.launcher = Launcher.instance()

        # Levels
        dummy = QtWidgets.QWidget()
        vbox = QtWidgets.QVBoxLayout()
        dummy.setLayout(vbox)
        self.use_level = TComboBox('Predict Level:', ['Level {}'.format(i) for i in self.LBLM.levels()])
        self.use_parent = TComboBox('Constrain Region:', ['None'])

        self.use_level.currentIndexChanged.connect(self.on_level_selected)
        self.use_parent.currentIndexChanged.connect(self.on_parent_selected)

        vbox.addWidget(self.use_level)
        vbox.addWidget(self.use_parent)

        self.LBLM.levelAdded.connect(self.on_levels_changed)
        self.LBLM.levelLoaded.connect(self.on_levels_changed)
        self.LBLM.levelRemoved.connect(self.on_levels_changed)

        self.LBLM.labelAdded.connect(self.on_label_added)
        self.LBLM.labelLoaded.connect(self.on_label_added)
        self.LBLM.labelNameChanged.connect(self.on_label_name_changed)
        self.LBLM.labelRemoved.connect(self.on_label_removed)

        self.addWidget(dummy)
        self.addWidget(HeaderLabel('Predict labels'))
        self.train_widget = TrainPredict()
        self.addWidget(self.train_widget)

        self.addWidget(HeaderLabel('Update annotations'))
        self.uncertainty_widget = Uncertainty()
        self.addWidget(self.uncertainty_widget)

        self.parent_labels = []
        self.levels = self.LBLM.levels()

    def on_level_selected(self, idx):
        if idx < 0:
            self.train_widget.select_level(None, None, -1)
            return
        self.train_widget.select_level(list(self.levels)[idx], None, -1)

        self.parent_labels = [(None, None)]
        self.use_parent.clear()
        self.use_parent.addItem('None')
        for i in range(len(self.levels)):
            if i == idx:
                continue
            prev_lvl = self.use_level.itemText(i)
            for label in self.LBLM.labels(self.levels[i]):
                self.parent_labels += [(self.levels[i], label)]
                self.use_parent.addItem('{}/{}'.format(prev_lvl, label.name))

    def on_parent_selected(self, idx):
        if idx > 0:
            self.train_widget.parent_level = self.parent_labels[idx][0]
            self.train_widget.parent_label = self.parent_labels[idx][1].idx
        else:
            self.train_widget.parent_level = None
            self.train_widget.parent_label = -1

    def on_levels_changed(self):
        self.levels = self.LBLM.levels()
        self.use_level.updateItems(['Level {}'.format(i) for i in self.levels])
        self.parent_level = None

    def on_label_added(self, level, dataset, label):
        current = self.use_level.currentIndex()
        if current >= 0 and list(self.levels)[current] != level:
            self.use_level.setCurrentIndex(self.levels.index(level))

    def on_label_name_changed(self, level, ds, label, name):
        current = self.use_level.currentIndex()
        if current >= 0 and list(self.levels)[current] != level:
            idx = -1
            for n, (lvl, lbl) in enumerate(self.parent_labels):
                if lvl == level and lbl.idx == label:
                    idx = n
                    break
            if idx > 0:
                prev_lvl = self.use_level.itemText(self.levels.index(lvl))
                self.use_parent.setItemText(idx, '{}/{}'.format(prev_lvl, name))

    def on_label_removed(self, level, ds, label):
        current = self.use_level.currentIndex()
        if current >= 0 and list(self.levels)[current] != level:
            idx = -1
            for n, (lvl, lbl) in enumerate(self.parent_labels):
                if lvl == level and lbl.idx == label:
                    idx = n
                    break
            if idx > 0:
                if self.train_widget.parent_level == level and \
                        self.train_widget.parent_label == label:
                    self.use_parent.setCurrentIndex(0)
                del self.parent_labels[idx]
                self.use_parent.removeItem(idx)




import os
import numpy as np
import six

rec_header_dtd = \
[
    ("nx", "i4"),              # Number of columns
    ("ny", "i4"),              # Number of rows
    ("nz", "i4"),              # Number of sections

    ("mode", "i4"),            # Types of pixels in the image. Values used by IMOD:
                               #  0 = unsigned or signed bytes depending on flag in imodFlags
                               #  1 = signed short integers (16 bits)
                               #  2 = float (32 bits)
                               #  3 = short * 2, (used for complex data)
                               #  4 = float * 2, (used for complex data)
                               #  6 = unsigned 16-bit integers (non-standard)
                               # 16 = unsigned char * 3 (for rgb data, non-standard)

    ("nxstart", "i4"),         # Starting point of sub-image (not used in IMOD)
    ("nystart", "i4"),
    ("nzstart", "i4"),

    ("mx", "i4"),              # Grid size in X, Y and Z
    ("my", "i4"),
    ("mz", "i4"),

    ("xlen", "f4"),            # Cell size; pixel spacing = xlen/mx, ylen/my, zlen/mz
    ("ylen", "f4"),
    ("zlen", "f4"),

    ("alpha", "f4"),           # Cell angles - ignored by IMOD
    ("beta", "f4"),
    ("gamma", "f4"),

    # These need to be set to 1, 2, and 3 for pixel spacing to be interpreted correctly
    ("mapc", "i4"),            # map column  1=x,2=y,3=z.
    ("mapr", "i4"),            # map row     1=x,2=y,3=z.
    ("maps", "i4"),            # map section 1=x,2=y,3=z.

    # These need to be set for proper scaling of data
    ("amin", "f4"),            # Minimum pixel value
    ("amax", "f4"),            # Maximum pixel value
    ("amean", "f4"),           # Mean pixel value

    ("ispg", "i4"),            # space group number (ignored by IMOD)
    ("next", "i4"),            # number of bytes in extended header (called nsymbt in MRC standard)
    ("creatid", "i2"),         # used to be an ID number, is 0 as of IMOD 4.2.23
    ("extra_data", "V30"),     # (not used, first two bytes should be 0)

    # These two values specify the structure of data in the extended header; their meaning depend on whether the
    # extended header has the Agard format, a series of 4-byte integers then real numbers, or has data
    # produced by SerialEM, a series of short integers. SerialEM stores a float as two shorts, s1 and s2, by:
    # value = (sign of s1)*(|s1|*256 + (|s2| modulo 256)) * 2**((sign of s2) * (|s2|/256))
    ("nint", "i2"),            # Number of integers per section (Agard format) or number of bytes per section (SerialEM format)
    ("nreal", "i2"),           # Number of reals per section (Agard format) or bit
                               # Number of reals per section (Agard format) or bit
                               # flags for which types of short data (SerialEM format):
                               # 1 = tilt angle * 100  (2 bytes)
                               # 2 = piece coordinates for montage  (6 bytes)
                               # 4 = Stage position * 25    (4 bytes)
                               # 8 = Magnification / 100 (2 bytes)
                               # 16 = Intensity * 25000  (2 bytes)
                               # 32 = Exposure dose in e-/A2, a float in 4 bytes
                               # 128, 512: Reserved for 4-byte items
                               # 64, 256, 1024: Reserved for 2-byte items
                               # If the number of bytes implied by these flags does
                               # not add up to the value in nint, then nint and nreal
                               # are interpreted as ints and reals per section

    ("extra_data2", "V20"),    # extra data (not used)
    ("imodStamp", "i4"),       # 1146047817 indicates that file was created by IMOD
    ("imodFlags", "i4"),       # Bit flags: 1 = bytes are stored as signed

    # Explanation of type of data
    ("idtype", "i2"),          # ( 0 = mono, 1 = tilt, 2 = tilts, 3 = lina, 4 = lins)
    ("lens", "i2"),
    ("nd1", "i2"),             # for idtype = 1, nd1 = axis (1, 2, or 3)
    ("nd2", "i2"),
    ("vd1", "i2"),             # vd1 = 100. * tilt increment
    ("vd2", "i2"),             # vd2 = 100. * starting angle

    # Current angles are used to rotate a model to match a new rotated image.  The three values in each set are
    # rotations about X, Y, and Z axes, applied in the order Z, Y, X.
    ("triangles", "f4", 6),    # 0,1,2 = original:  3,4,5 = current

    ("xorg", "f4"),            # Origin of image
    ("yorg", "f4"),
    ("zorg", "f4"),

    ("cmap", "S4"),            # Contains "MAP "
    ("stamp", "u1", 4),        # First two bytes have 17 and 17 for big-endian or 68 and 65 for little-endian

    ("rms", "f4"),             # RMS deviation of densities from mean density

    ("nlabl", "i4"),           # Number of labels with useful data
    ("labels", "S80", 10)      # 10 labels of 80 charactors
]


class MRC(object):

    def __init__(self, X, stats=None):
        self.header = self.header_dict = self.data = None
        self.yz_swapped = False

        if isinstance(X, six.string_types):
            self.read(X)
        else:
            # assuming X to be a numpy array
            self.parse(X, stats=stats)

    def __getitem__(self, item):
        if self.header_dict is not None and \
                item in self.header_dict:
            return self.header_dict[item]
        return None

    def keys(self):
        if self.header_dict is not None:
            return self.header_dict.keys()

    def values(self):
        if self.header_dict is not None:
            return self.header_dict.values()

    def items(self):
        if self.header_dict is not None:
            return self.header_dict.items()

    def parse(self, X, stats=None):
        if stats is not None:
            amin, amax, amean = stats
        else:
            amin = X.min(); amax = X.max(); amean = X.mean();

        dt = np.dtype(rec_header_dtd)
        imodFlags = 0
        if X.dtype in [np.uint8, np.int8]:
            mode = 0
            imodFlags = (X.dtype == np.int8)
        elif X.dtype == np.int16:
            mode = 1
        elif X.dtype == np.float32:
            mode = 2
        elif X.dtype == np.complex64:
            mode = 4
        elif X.dtype == np.uint16:
            mode = 6
        else:
            mode = 16

        values = (
            X.shape[2],              # Number of columns
            X.shape[1],              # Number of rows
            X.shape[0],              # Number of sections

            mode,                    # Types of pixels in the image. Values used by IMOD:
                                     #  0 = unsigned or signed bytes depending on flag in imodFlags
                                     #  1 = signed short integers (16 bits)
                                     #  2 = float (32 bits)
                                     #  3 = short * 2, (used for complex data)
                                     #  4 = float * 2, (used for complex data)
                                     #  6 = unsigned 16-bit integers (non-standard)
                                     # 16 = unsigned char * 3 (for rgb data, non-standard)

            0,                       # Starting point of sub-image (not used in IMOD)
            0,
            0,

            X.shape[2],              # Grid size in X, Y and Z
            X.shape[1],
            X.shape[0],

            X.shape[2],            # Cell size; pixel spacing = xlen/mx, ylen/my, zlen/mz
            X.shape[1],
            X.shape[0],

            90.0,           # Cell angles - ignored by IMOD
            90.0,
            90.0,

            # These need to be set to 1, 2, and 3 for pixel spacing to be interpreted correctly
            1,            # map column  1=x,2=y,3=z.
            2,            # map row     1=x,2=y,3=z.
            3,            # map section 1=x,2=y,3=z.

            # These need to be set for proper scaling of data
            amin,            # Minimum pixel value
            amax,            # Maximum pixel value
            amean,           # Mean pixel value

            1,            # space group number (ignored by IMOD)
            0,            # number of bytes in extended header (called nsymbt in MRC standard)
            0,         # used to be an ID number, is 0 as of IMOD 4.2.23
            "",     # (not used, first two bytes should be 0)

            # These two values specify the structure of data in the extended header; their meaning depend on whether the
            # extended header has the Agard format, a series of 4-byte integers then real numbers, or has data
            # produced by SerialEM, a series of short integers. SerialEM stores a float as two shorts, s1 and s2, by:
            # value = (sign of s1)*(|s1|*256 + (|s2| modulo 256)) * 2**((sign of s2) * (|s2|/256))
            0,            # Number of integers per section (Agard format) or number of bytes per section (SerialEM format)
            0,           # Number of reals per section (Agard format) or bit
                                       # Number of reals per section (Agard format) or bit
                                       # flags for which types of short data (SerialEM format):
                                       # 1 = tilt angle * 100  (2 bytes)
                                       # 2 = piece coordinates for montage  (6 bytes)
                                       # 4 = Stage position * 25    (4 bytes)
                                       # 8 = Magnification / 100 (2 bytes)
                                       # 16 = Intensity * 25000  (2 bytes)
                                       # 32 = Exposure dose in e-/A2, a float in 4 bytes
                                       # 128, 512: Reserved for 4-byte items
                                       # 64, 256, 1024: Reserved for 2-byte items
                                       # If the number of bytes implied by these flags does
                                       # not add up to the value in nint, then nint and nreal
                                       # are interpreted as ints and reals per section

            "",    # extra data (not used)
            0,       # 1146047817 indicates that file was created by IMOD
            imodFlags,       # Bit flags: 1 = bytes are stored as signed

            # Explanation of type of data
            0,          # ( 0 = mono, 1 = tilt, 2 = tilts, 3 = lina, 4 = lins)
            0,
            0,             # for idtype = 1, nd1 = axis (1, 2, or 3)
            0,
            0,             # vd1 = 100. * tilt increment
            0,             # vd2 = 100. * starting angle

            # Current angles are used to rotate a model to match a new rotated image.  The three values in each set are
            # rotations about X, Y, and Z axes, applied in the order Z, Y, X.
            [  0.,   0.,   0.,  90.,   0.,   0.],    # 0,1,2 = original:  3,4,5 = current

            0.,            # Origin of image
            X.shape[1] / 2.,
            0.,

            'MAP ',            # Contains "MAP "
            [68, 65,  0,  0],        # First two bytes have 17 and 17 for big-endian or 68 and 65 for little-endian

            0.0,             # RMS deviation of densities from mean density

            6,           # Number of labels with useful data
            [ 'tif2mrc: Converted to mrc format.                       21-Oct-15  13:02:25     ',
              'CCDERASER: Bad points replaced with interpolated values 21-Oct-15  13:03:42     ',
              'NEWSTACK: Images copied, transformed                    21-Oct-15  13:21:42     ',
              'TILT: Tomographic reconstruction                        30-Nov-15  12:29:39     ',
              'NEWSTACK: Images copied             , densities scaled  30-Nov-15  13:59:05     ',
              'clip: flipyz                                            30-Nov-15  13:59:10    ',
              '', '', '', '']      # 10 labels of 80 charactors
        )

        header = np.array(values, dtype=dt)

        header_dict = {}
        for name in header.dtype.names:
            header_dict[name] = header[name]

        self.header = header
        self.data = np.ascontiguousarray(X)
        self.header_dict = header_dict
        return self


    def read(self, filename):
        rec_header_dtype = np.dtype(rec_header_dtd)
        assert rec_header_dtype.itemsize == 1024

        fd = open(filename, 'rb')
        stats = os.stat(filename)

        header = np.fromfile(fd, dtype=rec_header_dtd, count=1)

        # Seek header
        #fd.seek(header.itemsize)
        if header['next'] > 0:
            fd.seek(header['next']) # ignore extended header

        mode = int(header['mode'])
        bo = "<" if header['stamp'][0, 0] == 68 and header['stamp'][0, 1] == 65 else "<" # BitOrder: little or big endian
        sign = "i1" if header['imodFlags'] == 1 else "u1" # signed or unsigned
                # 0     1     2     3     4     5     6     7     8     9     10    11    12    13    14    15    16
        dtype = [sign, "i2", "f",  "c4", "c8", None, "u2", None, None, None, None, None, None, None, None, None, "u1"][mode]
        dsize = [ 1,    2,    4,    4,    8,    0,    2,    0,    0,    0,    0,    0,    0,    0,    0,    0,    3][mode]

        # data dimensions
        nx, ny, nz = int(header['nx']), int(header['ny']), int(header['nz'])
        img_size = nx * ny * nz * dsize
        img_str = fd.read(img_size)
        dtype = bo + dtype

        # Make sure that we have readed the whole file
        assert not fd.read(), "Error loading the file"
        #assert stats.st_size == header.itemsize + img_size

        fd.close()

        if mode == 16:
            data = np.ndarray(shape=(nx, ny, nz, 3), dtype=dtype, buffer=img_str, order='F')
        else:
            data = np.ndarray(shape=(nx, ny, nz), dtype=dtype, buffer=img_str, order='F')

        data = np.swapaxes(data, 0, 2)

        header_dict = {}
        for name in header.dtype.names:
            header_dict[name] = header[name][0] if len(header[name]) == 1 else header[name]

        self.header = header
        self.data = np.ascontiguousarray(data)
        self.header_dict = header_dict
        return self

    def save(self, filename):
        fd = open(filename, 'wb')
        fd.write(self.header.data)
        data = self.data
        data = np.swapaxes(data, 0, 2)
        if self.yz_swapped:
            data = np.swapaxes(data, 1, 2)
        data = np.asfortranarray(data)
        fd.write(data.data)
        fd.close()




import numpy as np
from ..qt_compat import QtGui, QtCore, QtWidgets

from collections import OrderedDict

from .base import Plugin
from ..widgets import HLayer, HeaderLabel, SubHeaderLabel
from ..core import DataModel, LayerManager, Launcher
from .. import actions as ac


class Layers(QtWidgets.QWidget):

    name = 'Layers'

    def __init__(self, parent=None):
        super(Layers, self).__init__(parent=parent)

        self.setLayout(QtWidgets.QVBoxLayout())

        self.levels = OrderedDict()
        self.DM = DataModel.instance()
        self.LM = LayerManager.instance()
        self.LM.added.connect(self.on_layer_added)
        self.LM.removed.connect(self.on_layer_removed)
        self.LM.toggled.connect(self.setVisibility)

    def __del__(self):
        self.levels.clear()

    def on_layer_added(self, name, level):
        name, level = str(name), str(level)
        if level not in self.levels:
            self.levels[level] = OrderedDict()

        lobj = self.LM.get(name, level)
        layer = HLayer(name, level, current=int(lobj.alpha*100), visible=lobj.visible)
        layer.valueChanged.connect(self.on_value)
        layer.toggled.connect(self.on_toggled)
        layer.export.connect(self.on_tiff_export)
        self.levels[level][name] = layer

        self.redraw_layers()

    def redraw_layers(self):
        for i in reversed(range(self.layout().count())):
            self.layout().itemAt(i).widget().setParent(None)

        for level in self.LM.levels():
            header = SubHeaderLabel(level)
            self.layout().addWidget(header)
            for layer in self.LM.layer_names(level=level):
                self.layout().addWidget(self.levels[level][layer])
        self.update()

    def on_layer_removed(self, name, level):
        del self.levels[level][name]
        self.redraw_layers()

    def on_value(self, name, level, val):
        self.LM.setOpacity(name, level, val / 100.)

    def on_toggled(self, name, level, bol):
        self.LM.setVisible(name, level, bol)
        self.LM.update()

    def setVisibility(self, name, level, bol):
        self.levels[level][name].setChecked(bol)

    def on_tiff_export(self, name, level):
        path = str(QtWidgets.QFileDialog.getSaveFileName(self, filter='*.rec'))
        if path is None or path == '':
            return
        if not path.endswith('.rec'):
            path += '.rec'
        layer = self.LM.get(name, level)
        Launcher.instance().run(ac.export_tiff_layer, data=layer.data, level=level,
                                path=path, cb=None, caption='Exporting layer %s..' % name)



from __future__ import unicode_literals    # at top of module

import numpy as np
import h5py as h5

from ..qt_compat import QtGui, QtCore, QtWidgets
from matplotlib.colors import ListedColormap

import os
import logging as log
from collections import OrderedDict

from .base import Plugin
from ..widgets import HWidgets, HEditLabel, HeaderLabel, PLineEdit, \
                      TComboBox, SubHeaderLabel, BLabel, RoundedWidget,\
                      ComboDialog, RCheckBox, HSize3D, SComboBox, ActionButton
from ..core import DataModel, LabelManager, LayerManager, Launcher
from .. import actions as ac
from six import iteritems


class AnnotationLayerLabel(QtWidgets.QWidget):

    def __init__(self, text, parent=None):
        super(AnnotationLayerLabel, self).__init__(parent=parent)

        hbox = QtWidgets.QHBoxLayout()
        self.setLayout(hbox)

        spacing = QtWidgets.QPushButton('X')
        spacing.setMaximumWidth(25)
        spacing.setMaximumHeight(30)
        spacing.setMinimumHeight(30)
        spacing.setMinimumWidth(25)
        spacing.setStyleSheet('border-top-right-radius: 0px;'
                              'border-bottom-right-radius: 0px;'
                              'border-right: 2px solid red;')
        self.deleted = spacing.clicked

        hbox.addWidget(spacing)

        label = QtWidgets.QLabel(text)
        label.setMinimumHeight(30)
        label.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter)
        label.setStyleSheet('background-color: #6DC7C7; color: #006161;'
                            'padding-left: 10px;'
                            'border-top: 1px solid #fefefe;'
                            'border-bottom: 1px solid #fefefe;'
                            'font-weight: bold; font-size: 13px;')

        self.layout().setContentsMargins(0,0,0,0)
#        self.layout().setMargin(0)
        self.layout().setSpacing(0)

        hbox.addWidget(label, 1)

        self.chk = QtWidgets.QCheckBox()
        self.chk.setMinimumHeight(30)
        self.chk.setMaximumWidth(25)
        self.chk.setMaximumHeight(30)
        self.chk.setMinimumWidth(25)
        self.chk.setStyleSheet('background-color: #6DC7C7;'
                               'border-top-right-radius: 3px;'
                               'border-bottom-right-radius: 3px;'
                               'border: 1px solid #fefefe;'
                               'border-left: none')
        hbox.addWidget(self.chk)
        self.stateChanged = self.chk.stateChanged

    def setChecked(self, bol):
        self.chk.setChecked(bol)


class Level(RoundedWidget):

    parent_requested = QtCore.pyqtSignal(int, int)
    level_toggled = QtCore.pyqtSignal(int, bool)
    erase = QtCore.pyqtSignal(int)

    def __init__(self, name, level, ds, parent=None):
        super(Level, self).__init__(parent=parent, color=None, bg='#cde5e5')

        self.DM = DataModel.instance()
        self.LM = LayerManager.instance()
        self.LBLM = LabelManager.instance()

        self.name = name
        self.level = level
        self.dataset = ds

        self.setLayout(QtWidgets.QVBoxLayout())
        self.setContentsMargins(0,0,0,0)

        self.header = AnnotationLayerLabel(name)
        self.layout().addWidget(self.header)
        self.btn_addlabel = ActionButton('Add Label')
        self.btn_erase = ActionButton('Erase')
        self.btn_erase.setCheckable(True)
        self.layout().addWidget(HWidgets(None, self.btn_erase, self.btn_addlabel,
                                stretch=[1,0,0]))

        self.labels = OrderedDict()

        self.btn_addlabel.clicked.connect(self.on_add_label)
        self.btn_erase.clicked.connect(self.on_erase)
        self.header.deleted.connect(self.on_remove_level)

        self.LM.toggled.connect(self.toggle_level)
        self.header.stateChanged.connect(self.on_level_toggled)

    def num_labels(self):
        return len(self.labels)

    def toggle_level(self, name, level, val):
        if level == 'Annotations' and name == self.name:
            self.header.setChecked(val)

    def on_level_toggled(self, bol):
        self.level_toggled.emit(self.level, bol)

    def add_label(self, label, name):
        wlabel = HEditLabel(name, label)
        wlabel.selected.connect(self.on_selected)
        wlabel.removed.connect(self.on_remove_label)
        wlabel.nameChanged.connect(self.on_nameChanged)
        wlabel.colorChanged.connect(self.on_colorChanged)
        wlabel.visibilityChanged.connect(self.on_visibilityChanged)
        wlabel.parentRequested.connect(self.on_parentRequested)

        self.labels[label] = wlabel
        self.layout().addWidget(wlabel)

    def load_label(self, label, name, color, visible, parent_level, parent_label):

        wlabel = HEditLabel(name, label)
        wlabel.setName(name)
        wlabel.setColor(color)
        wlabel.setChecked(visible)

        wlabel.selected.connect(self.on_selected)
        wlabel.removed.connect(self.on_remove_label)
        wlabel.nameChanged.connect(self.on_nameChanged)
        wlabel.colorChanged.connect(self.on_colorChanged)
        wlabel.visibilityChanged.connect(self.on_visibilityChanged)
        wlabel.parentRequested.connect(self.on_parentRequested)

        self.labels[label] = wlabel
        self.layout().addWidget(wlabel)
        self.set_parent(label, parent_level, parent_label)

    def remove_label(self, label):
        if label in self.labels:
            self.labels[label].setParent(None)
            del self.labels[label]

    def setSelected(self, label):
        for k, lbl in iteritems(self.labels):
            lbl.setSelected(k == label)
        self.btn_erase.setChecked(label is not None and label < 0)

    def on_remove_level(self):
        self.LBLM.removeLevel(self.level)

    def on_add_label(self):
        self.LBLM.addLabel(self.level)

    def on_remove_label(self, label):
        self.LBLM.removeLabel(self.level, label)

    def on_erase(self):
        self.erase.emit(self.level)

    def on_selected(self, label):
        self.LBLM.selectLabel(self.level, label)

    def on_nameChanged(self, label, name):
        self.LBLM.changeLabelName(self.level, label, str(name))

    def on_colorChanged(self, label, color):
        self.LBLM.changeLabelColor(self.level, label, str(color))

    def on_visibilityChanged(self, label, visible):
        self.LBLM.changeLabelVisibility(self.level, label, visible)

    def on_parentRequested(self, label):
        self.parent_requested.emit(self.level, label)

    def set_parent(self, label, parent_level, parent_label):
        if parent_label < 0:
            color = None
        else:
            color = self.LBLM.get(parent_level, parent_label).color
        self.labels[label].setParentColor(color)

    def update_parents(self, parent_level, parent_label):
        for label in self.labels.keys():
            obj = self.LBLM.get(self.level, label)
            if obj.parent_level == parent_level and \
                    obj.parent_label == parent_label:
                self.set_parent(label, parent_label)

    def visible(self, bol=None):
        if bol is None:
            return self.LM.get(self.name, 'Annotations').visible
        else:
            self.LM.setVisible(self.name, 'Annotations', bol)


class Annotations(Plugin):

    name = 'Annotations'

    def __init__(self):
        super(Annotations, self).__init__(ptype=Plugin.Plugin)

        # Relayout
        QtWidgets.QWidget().setLayout(self.layout)
        vbox2 = QtWidgets.QVBoxLayout()
        vbox2.setContentsMargins(0, 10, 0, 0)
        self.setLayout(vbox2)

        self.labels = {}
        self.DM = DataModel.instance()
        self.LM = LayerManager.instance()
        self.LBLM = LabelManager.instance()
        self.launcher = Launcher.instance()

        self.levels = OrderedDict()

        #######
        vbox2.addWidget(HeaderLabel('Annotations'))

        dummy = QtWidgets.QWidget()
        vbox = QtWidgets.QVBoxLayout()
        dummy.setLayout(vbox)

        vbox2.addWidget(dummy, 1)

        vbox.addWidget(SubHeaderLabel('Annotation Levels'))

        self.level_background = PLineEdit(-1, parse=int)
        self.level_background.setMaximumWidth(40)
        self.btn_loadlevel = ActionButton('Load Level')
        self.btn_addlevel = ActionButton('Add Level')
        vbox.addWidget(HWidgets('Bg:', self.level_background,
                                self.btn_loadlevel, None, self.btn_addlevel,
                                stretch=[0, 0, 0, 1,0]))

        groupbox = QtWidgets.QWidget()
        self.form = QtWidgets.QFormLayout()
        self.form.setContentsMargins(0, 0, 0, 0)
        groupbox.setLayout(self.form)

        scroll = QtWidgets.QScrollArea()
        scroll.setWidget(groupbox)
        scroll.setWidgetResizable(True)
        vbox.addWidget(scroll, 1)

        ###############################################
        vbox.addWidget(SubHeaderLabel('Refine Annotation Label'))
        self.combo_refine = TComboBox('', ['dilation', 'erosion', 'opening',
                                           'closing', 'fill_holes'])
        self.refine_radius = PLineEdit(1, parse=int)
        self.refine_btn = ActionButton('Refine')
        self.slide_combo = SComboBox()
        self.slide_combo.addItem('This slice')
        self.slide_combo.addItem('All slices (2D)')
        self.slide_combo.addItem('Whole volume (3D)')
        vbox.addWidget(HWidgets(self.combo_refine, 'Radius:', self.refine_radius,
                                None, self.slide_combo, stretch=[1, 0, 0, 1, 0]))
        vbox.addWidget(HWidgets(None, self.refine_btn, stretch=[1, 0]))

        ### Signals
        self.btn_addlevel.clicked.connect(self.LBLM.addLevel)
        self.btn_loadlevel.clicked.connect(self.on_load_external_level)
        #self.previous_tool.clicked.connect(self.on_load_previous_tool_level)
        self.refine_btn.clicked.connect(self.refine_level)

        self.LBLM.levelAdded.connect(self.on_add_level)
        self.LBLM.levelLoaded.connect(self.on_load_level)
        self.LBLM.levelRemoved.connect(self.on_remove_level)
        self.LBLM.saveLevel.connect(self.save_labels)

        self.LBLM.labelAdded.connect(self.on_add_label)
        self.LBLM.labelRemoved.connect(self.on_remove_label)
        self.LBLM.labelLoaded.connect(self.on_load_label)
        self.LBLM.labelSelected.connect(self.on_select_label)
        self.LBLM.labelNameChanged.connect(self.on_label_name)
        self.LBLM.labelColorChanged.connect(self.on_label_color)
        self.LBLM.labelVisibilityChanged.connect(self.on_label_visibility)
        self.LBLM.labelParentChanged.connect(self.on_label_parent_changed)

    def refine_level(self):
        if self.DM.gtselected is None:
            QtWidgets.QMessageBox.critical(self, 'Error', 'No label selected')
            return

        level = self.DM.gtselected['level']
        label = self.DM.gtselected['label']
        refine_method = self.combo_refine.currentText()
        refine_radius = self.refine_radius.value()
        rslide = [self.DM.current_idx, '2D', '3D'][self.slide_combo.currentIndex()]

        self.launcher.run(ac.refine_label, data=level, label=label, slide=rslide,
                          method=refine_method, radius=refine_radius,
                          caption='Refining annotations..', cb=self.on_refined)

    def on_refined(self, params):
        if params is None:
            return
        indexes, values = params
        if len(indexes) == 0 or len(values) == 0:
            return

        zmin, ymin, xmin = indexes.min(0)
        zmax, ymax, xmax = indexes.max(0)

        indexes[:, 0] -= zmin
        indexes[:, 1] -= ymin
        indexes[:, 2] -= xmin

        slice_z = slice(zmin, zmax+1)
        slice_y = slice(ymin, ymax+1)
        slice_x = slice(xmin, xmax+1)

        level = self.DM.gtselected['level']
        self.DM.last_changes = (level, (slice_z, slice_y, slice_x),
                                indexes, values, False)
        self.LM.update()

    def on_add_level(self, idx, ds):
        assert idx not in self.levels

        success = self.create_level_ds(ds)
        if not success:
            errmsg = 'Dataset "{}" already exists'.format(ds)
            QtWidgets.QMessageBox.critical(self, 'Error', errmsg)
            return
        else:
            self.DM.set_attrs(ds, dict(active=True, levelid=idx))

        name = u'Level {}'.format(idx)
        level = Level(name, idx, ds)
        level.level_toggled.connect(self.on_level_toggled)
        level.erase.connect(self.on_erase_clicked)
        level.parent_requested.connect(self.on_request_parent)

        self.levels[idx] = level
        self.form.addRow(level)

        cmap = ListedColormap(['#000000'])
        self.LM.addLayer(ds, name, 'Annotations', cmap=cmap,
                         background=-1, visible=False, vmin=0, vmax=1)

        self.save_labels(idx)
        log.info('+ Ready.')
        return ds

    def on_load_level(self, idx, ds):
        assert idx not in self.levels, "Error, level ID already exists"

        name = u'Level {}'.format(idx)
        level = Level(name, idx, ds)
        level.level_toggled.connect(self.on_level_toggled)
        level.erase.connect(self.on_erase_clicked)
        level.parent_requested.connect(self.on_request_parent)

        self.levels[idx] = level
        self.form.addRow(level)

        cmap = ListedColormap(['#000000'])
        self.LM.addLayer(ds, name, 'Annotations', cmap=cmap,
                         background=-1, visible=False, vmin=0, vmax=1)
        log.info('+ Loading [Level {}] -- "{}"'.format(idx, ds))
        return ds

    def on_load_external_level(self):
        self.launcher.setup("Importing annotation level")
        path = QtWidgets.QFileDialog.getOpenFileName(self, "Select input source",
                                                 filter='*.h5 *.hdf5 *.h5.bak')
        if path is None or len(path) == 0:
            self.launcher.cleanup()
            return

        dataset = None

        p1 = os.path.basename(path)
        p2 = os.path.dirname(path)
        p3 = os.path.basename(p2)
        p4 = os.path.realpath(os.path.dirname(p2))
        p5 = os.path.realpath(self.DM.wspath)

        if p4 == p5 and p3 == 'annotations' and p1.endswith('.h5'):
            log.info('+ Attempting to re-activate a previous level')
            dataset = 'annotations/{}'.format(p1[:-3])
            attrs = self.DM.attrs(dataset)

            if 'levelid' in attrs and 'label' in attrs and 'parent_levels' in attrs \
                    and 'parent_labels' in attrs and 'colors' in attrs \
                    and 'names' in attrs and 'visible' in attrs and 'active' in attrs \
                    and (len(attrs['label']) == len(attrs['parent_levels']) \
                         == len(attrs['colors']) == len(attrs['parent_labels']) \
                         == len(attrs['names']) == len(attrs['visible'])):
                if attrs['active'] == False:
                    levels = self.LBLM.levels()
                    if attrs['levelid'] in levels:
                        levelid = max(levels) + 1
                        self.DM.set_attrs(dataset, dict(levelid=levelid))
                    else:
                        levelid = attrs['levelid']
                    self.DM.set_attrs(dataset, dict(active=True))
                    last_level, ds = self.LBLM.loadLevel(levelid, dataset)
                    labels = self.DM.attr(dataset, 'label')
                else:
                    errmsg = 'Selected level is already loaded'
                    QtWidgets.QMessageBox.critical(self, 'Error', errmsg)
                    return self.launcher.cleanup()
            else:
                log.info('    * Not possible to re-activate')


        if path.endswith('.h5') or path.endswith('.h5.bak'):
            av = self.DM.available_hdf5_datasets(path, dtype=np.int16)
            if len(av) == 0:
                self.launcher.show_error('No annotation level found')
                return self.launcher.cleanup()
            selected, accepted = ComboDialog.getOption(av)
            if accepted == QtWidgets.QDialog.Rejected:
                return
            dataset = selected

            background = self.level_background.value()

            with h5.File(path, 'r') as f:
                shape = f[dataset].shape
            dshape = self.DM.data_shape
            rshape = self.DM.region_shape()

            if shape != dshape and shape != rshape:
                errmsg = 'Annotation shape {} do not match with either loaded dataset {}' \
                         ' or selected ROI {}'.format(shape, dshape, rshape)
                self.launcher.show_error(errmsg)
                self.launcher.cleanup()
                return

            last_level, ds = self.LBLM.addLevel()

            with h5.File(path, 'r') as f:
                data = f[dataset][:]

            if background != -1:
                data[data == background] = -1
            if data.dtype != np.int32:
                data = data.astype(np.int32)

            log.info('+ Writing labels to disk')
            if shape == dshape:
                self.DM.write_dataset(ds, data)
            else:
                self.DM.write_slices(ds, data)

            labels = set(np.unique(data)) - set([-1])

        log.info('+ Populating widgets')
        for label in labels:
            self.LBLM.loadLabel(last_level, label, 'Label {}'.format(label),
                                '#000000', True, -1, -1)
        self.on_level_toggled(last_level, True)
        self.save_labels(last_level)
        log.info('+ Ready.')
        self.launcher.cleanup()

    def on_remove_level(self, idx, ds):
        visible = self.levels[idx].visible()

        log.info('+ Removing [Level {}]'.format(idx))
        self.LM.remove(self.levels[idx].name, 'Annotations')

        if self.DM.gtselected is not None and self.DM.gtselected['levelidx'] == idx:
            self.DM.gtselected = None

        log.info('+ Clearing annotations [Level {}]'.format(idx))
        self.remove_level_ds(idx, ds)

        log.info('+ Removing widgets')
        self.levels[idx].setParent(None)
        del self.levels[idx]

        if visible:
            self.LM.update()

        log.info('+ Ready.')

    def on_level_toggled(self, level, bol):
        for lobj in self.levels.values():
            if lobj.level == level and not bol:
                lobj.setSelected(None)

        if not bol and self.DM.gtselected is not None and \
                self.DM.gtselected['levelidx'] == level:
            self.DM.gtselected = None

        self.levels[level].visible(bol)
        self.LM.update()

    def on_erase_clicked(self, level):
        self.on_select_label(level, self.LBLM.dataset(level), -1)

    def on_load_label(self, level, dataset, label, name, color, visible,
                      parent_level, parent_label):
        log.info('  * Loading annotation [{}] for [Level {}]'.format(name, level))
        self.levels[level].load_label(label, name, color, visible, parent_level, parent_label)
        self.update_level_colormap(level)

    def on_add_label(self, level, dataset, label, name):
        self.levels[level].add_label(label, name)
        self.save_labels(level)
        self.update_level_colormap(level)

        if self.levels[level].visible():
            self.LM.update()
        log.info('+ Ready.')

    def on_remove_label(self, level, dataset, labelobj):
        label = labelobj.idx
        label_name = labelobj.name

        if self.DM.gtselected is not None:
            slevel = self.DM.gtselected['levelidx']
            slabel = self.DM.gtselected['label']
            if slevel == level:
                if slabel == label:
                    self.DM.gtselected = None

        self.launcher.setup('Removing label [{}] form [Level {}]'
                            .format(label_name, level))
        log.info('+ Loading annotation data into memory')
        data = self.DM.load_ds(dataset)
        data[data == label] = -1

        log.info('+ Writing changes to disk')
        self.DM.write_dataset(dataset, data)

        log.info('+ Removing widgets')
        self.levels[level].remove_label(label)
        self.save_labels(level)
        self.update_level_colormap(level)

        self.LM.update()

        self.launcher.cleanup()

    def on_select_label(self, level, dataset, label):
        for idx, lobj in iteritems(self.levels):
            if idx == level:
                lobj.setSelected(label)
                self.LM.setVisible(lobj.name, 'Annotations', True)
            else:
                lobj.setSelected(None)

        self.select_label(level, dataset, label)
        self.LM.update()

    def select_label(self, level, dataset, label):

        if level is None:
            self.DM.gtselected = None
        elif  self.DM.gtselected is None or self.DM.gtselected['levelidx'] != level:
            parent_level = parent_label = None
            if label >= 0:
                current = self.LBLM.get(level, label)
                if current.parent_level is not None and current.parent_label >= 0:
                    parent_level = current.parent_level
                    parent_label = current.parent_label

            self.DM.gtselected = {
                'levelidx' : level,
                'level' : self.LBLM.dataset(level),
                'label' : label,
                'parent_level' : parent_level,
                'parent_label' : parent_label
            }
            lobj = self.LM.get(self.levels[level].name, 'Annotations')
            lobj.data = self.LBLM.dataset(level)
        elif self.DM.gtselected['levelidx'] == level:
            parent_level = parent_label = None
            if label >= 0:
                current = self.LBLM.get(level, label)
                if current.parent_level is not None and current.parent_label >= 0:
                    parent_level = current.parent_level
                    parent_label = current.parent_label
            self.DM.gtselected['label'] = label
            self.DM.gtselected['parent_level'] = parent_level
            self.DM.gtselected['parent_label'] = parent_label

        log.info('+ Ready.')

    def on_label_name(self, level, label, name):
        self.save_labels(level)

    def on_label_color(self, level, label, name):
        self.save_labels(level)
        self.update_level_colormap(level)
        for level in self.levels.values():
            level.update_parents(level, label)
        self.LM.update()

    def on_label_visibility(self, level, label, name):
        self.save_labels(level)

    def update_level_colormap(self, level):
        colors = self.LBLM.colors(level)
        idxs = self.LBLM.idxs(level)
        maxid = 0 if len(idxs) == 0 else max(idxs)+1
        cmaplist = ['#000000'] * maxid
        for idx, color in zip(idxs, colors):
            cmaplist[idx] = color
        newcmap = ListedColormap(cmaplist)
        layer = self.LM.get(self.levels[level].name, 'Annotations')
        layer.cmap = newcmap
        layer.vmin = 0
        layer.vmax = maxid

    def save_labels(self, level):
        log.info('+ Updating labels for [Level {}]'.format(level))
        dataset = self.levels[level].dataset
        self.DM.set_attrs(dataset, dict(label   = self.LBLM.idxs(level),
                                        names   = self.LBLM.names(level),
                                        colors  = self.LBLM.colors(level),
                                        visible = list(map(int, self.LBLM.visibility(level))),
                                        parent_levels = self.LBLM.parent_levels(level),
                                        parent_labels = self.LBLM.parent_labels(level)))

    def create_level_ds(self, ds):
        log.info('### Creating annotations "{}" ###'.format(ds))
        return ds, self.DM.create_empty_dataset(ds, self.DM.data_shape,
                                                dtype=np.int16)

    def remove_level_ds(self, idx, ds):
        quit_msg = "Do you want to permanently remove the data from [Level {}]?".format(idx)
        wipe = QtWidgets.QMessageBox.question(self, 'Message', quit_msg,
                                          QtWidgets.QMessageBox.Yes,
                                          QtWidgets.QMessageBox.No)
        if wipe == QtWidgets.QMessageBox.Yes:
            log.info('### Removing [Level {}] annotations ###'.format(idx))
            self.DM.remove_dataset(ds)
        else:
            log.info('### Disabling [Level {}] annotations ###'.format(idx))
            self.DM.set_attrs(ds, dict(active=False))

    def on_request_parent(self, level, label):
        if level == 0:
            return

        current_index = list(self.levels.keys()).index(level)
        if current_index > 0:
            parent_level = list(self.levels.keys())[current_index-1]
            dataset = self.LBLM.dataset(level)
            parent_dataset = self.LBLM.dataset(parent_level)
        else:
            return

        labels = self.LBLM.labels(parent_level)
        if len(labels) == 0:
            return
        
        options = ['None'] + [u'Level {}/{}'.format(parent_level, l.name) for l in labels]
        option, result = ComboDialog.getOptionIdx(options)

        if not result:
            return

        if option == 0:
            selected = None
            parent_label = -1
        else:
            selected = labels[option - 1]
            parent_label = selected.idx

        if selected is None:
            if self.DM.gtselected is not None \
                    and self.DM.gtselected['levelidx'] == level \
                    and self.DM.gtselected['label'] == label:
                self.DM.gtselected['parent_level'] = None
            self.LBLM.setLabelParent(level, label, parent_level, parent_label)
            #self.levels[level].set_parent(label, parent_label)
            self.save_labels(level)
            return

        self.launcher.setup('Assigning parent to [{}] of [Level {}]'
                            .format(str(self.LBLM.labels(level)[label].name,'utf-8'), level))

        dparent = self.DM.load_ds(parent_dataset)
        dcurrent = self.DM.load_ds(dataset)

        mparent = (dparent == parent_label)
        mcurrent = (dcurrent == label)
        mintersection = mcurrent & mparent
        if not (mcurrent == mintersection).all():
            err_msg = "Some annotations of the label will be lost, do you wish to continue?"
            ans = QtWidgets.QMessageBox.question(self, "Data will be lost", err_msg,
                                             QtWidgets.QMessageBox.Yes,
                                             QtWidgets.QMessageBox.No)
            if ans == QtWidgets.QMessageBox.No:
                return

            dcurrent[mcurrent] = -1
            dcurrent[mintersection] = label
            self.DM.write_dataset(dataset, dcurrent)

            slices = slice(None), slice(None), slice(None)
            indexes = np.column_stack(np.where(mcurrent))
            self.DM.last_changes = (dataset, slices, indexes, label)

        self.LBLM.setLabelParent(level, label, parent_level, parent_label)
        self.launcher.cleanup()

    def on_label_parent_changed(self, level, dataset, label, parent_level, parent_label):
        if self.DM.gtselected is not None \
                and self.DM.gtselected['levelidx'] == level \
                and self.DM.gtselected['label'] == label:

            self.DM.gtselected['parent_level'] = parent_level
            self.DM.gtselected['parent_label'] = parent_label

        self.levels[level].set_parent(label, parent_level, parent_label)
        self.save_labels(level)
        self.LM.update()
